import unittest

'''
Prgram 명: Glance of Global countries
Author: jeongmin Kang
Mail: jarvisNim@gmail.com
글로벌 경쟁국들의 투자를 위한 경제지표(거시/미시) 부문 에측 전망치(Outlook) 제공
- OECD
- IMF
* investing.com/calendar 포함
History
20231111  Create
20240130  Alpha 테이블 구성, 그래프 추가 
'''

import sys, os
utils_dir = os.getcwd() + '/batch/utils'
sys.path.append(utils_dir)

from settings import *
from economics_db import make_alpha


'''
0. 공통영역 설정
'''

import requests
import yfinance as yf
import pandas_ta as ta
from bs4 import BeautifulSoup as bs


# logging
logger.warning(sys.argv[0] + ' :: ' + str(datetime.today()))
logger2.info(sys.argv[0] + ' :: ' + str(datetime.today()))

# 3개월단위로 순차적으로 읽어오는 경우의 시작/종료 일자 셋팅
to_date_2 = pd.to_datetime(to_date2)
three_month_days = relativedelta(weeks=12)
from_date = (to_date_2 - three_month_days).date()
to_date_2 = to_date_2.date()

database = 'Economics.db'
db_file = 'database/' + database
conn, engine = create_connection(db_file)

month_terms = [0, 3, 6, 12, 18, 24]  # 현재와 n 개월 전망치 값


def get_copper():
    copper = fred.get_series(series_id='PCOPPUSDM', observation_start=from_date_MT)
    result = (copper[-1] - copper[0]) / copper[0]
    return result + 1


def wgbonds_10y(country_sign):
    if country_sign == 'KR':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/south-korea/10-years/"
    elif country_sign == 'JP':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/japan/10-years/"
    elif country_sign == 'CN':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/china/10-years/"
    elif country_sign == 'BR':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/brazil/10-years/"
    elif country_sign == 'DE':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/germany/10-years/"
    elif country_sign == 'SG':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/singapore/10-years/"
    elif country_sign == 'IN':
        url = "https://www.worldgovernmentbonds.com/bond-forecast/india/10-years/"           
    else:
        url =""

    page = requests.get(url)
    soup = bs(page.text, "html.parser")
    # print(soup)
    tables = soup.find_all('table')
    # print(tables)
    # 멀티 헤더의 첫번째 헤더 제거
    y10 = pd.read_html(str(tables))[0]
    column_to_remove = y10.columns[0]

    index = y10.columns
    second_indices = [idx[1] for idx in index]
    logger2.info(y10)
    logger2.info('')
    result = pd.DataFrame(columns=['Date', 'Yield_10y'])
    data = []

    for i in range(len(index)):
        my_dict = {}
        try:
            _date = datetime.strptime(second_indices[i], '%d %b %Y')
        except:  # 'Jun 2024'
            _date = datetime.strptime(second_indices[i], '%b %Y')
        finally:
            date = _date.strftime('%Y-%m-%d')

        string = y10.iloc[0,i]
        idx = string.find("%")  # '%' 문자의 인덱스 찾기
        if idx != -1:  # '%' 문자가 발견되었다면
            yield_10y = string[:idx]  # '%' 문자의 앞 글자 추출
        else:
            logger.error(">>> global_.py 에서 '%'' 문자를 찾을 수 없습니다.")

        my_dict['Date'] = date
        my_dict['Yield_10y'] = yield_10y
        # print(my_dict)        
        data.append(my_dict)

    result = pd.DataFrame().from_records(data)
    result['Date'] = pd.to_datetime(result['Date']).dt.date
    result['Yield_10y'] = pd.to_numeric(result['Yield_10y'])
    result['Date'].reset_index()
    print(result)

    logger2.info('')
    logger2.info(f'##### World Government Bonds: {country_sign} #####')
    logger2.info('10Year bond yield: \n' + str(result))

    return result


'''
1. Economics Area
1.1 Leading Indicators OECD: CLI (Composite leading indicator)
'''

def oecd_cli():

    # Economics database 에서 쿼리후 시작하는 루틴
    M_table = 'OECD'
    # M_countries = "('OECD - Total', 'United States', 'Korea', 'Japan', 'China (People's Republic of)', 'Germany')"
    M_countries = ['United States', 'Korea', 'Japan', 'Germany', 'China', 'OECD - Total']

    for country in M_countries:

        M_query = f"SELECT * from {M_table} WHERE Country like '{country}%' ORDER by Variable ASC, Time ASC"

        try:
            df = pd.read_sql_query(M_query, conn)
            df = df.sort_values(['Variable', 'Time'], ascending=True).reset_index(drop=True)
            # logger2.info(df.head(5))
        except Exception as e:
            logger.error(' >>> Exception: {}'.format(e))

        events = df['Variable'].unique()

        # 전체 그림의 크기를 설정
        plt.figure(figsize=(16, 3*len(events)))
        for i, event in enumerate(events):
            result = df[df['Variable'] == event]        
            if result.empty:
                continue
            result.dropna()
            country = result.iloc[0]['Country']
            result['Time'] = pd.to_datetime(result['Time']).dt.year
            result['Variable'] = result['Variable']
            plt.subplot(len(events), 1, i + 1)
            plt.plot(result['Time'], result['Value'])
            max_val = max(result['Value'])
            min_val = min(result['Value'])
            if (max_val > 0) and (min_val < 0):       # 시각효과     
                plt.axhline(y=0, linestyle='--', color='red', linewidth=1)
            plt.title(f"OECD Outlook {country}: {event}")
            plt.grid()
            plt.xlabel('Time')
            plt.ylabel('Value')

        plt.tight_layout()  # 서브플롯 간 간격 조절
        plt.savefig(reports_dir + f'/global_0100_{country}.png')

    return df


def fed_cli():
    CLI_OECD_Total = fred.get_series(series_id='OECDLOLITONOSTSAM', observation_start=from_date_MT)
    CLI_OECD_Total_Plus_Six = fred.get_series(series_id='ONMLOLITONOSTSAM', observation_start=from_date_MT)
    CLI_Usa = fred.get_series(series_id='USALOLITONOSTSAM', observation_start=from_date_MT)
    CLI_Korea = fred.get_series(series_id='KORLOLITONOSTSAM', observation_start=from_date_MT)
    CLI_China = fred.get_series(series_id='CHNLOLITONOSTSAM', observation_start=from_date_MT)
    CLI_Germany = fred.get_series(series_id='DEULOLITOTRSTSAM', observation_start=from_date_MT)

    logger2.info('##### Leading indicators: CLI (Composite leading indicator) #####')
    logger2.info('CLI_OECD_Total: \n' + str(CLI_OECD_Total[-3:]))
    logger2.info('')
    logger2.info('CLI_OECD_Total_Plus_Six: \n' + str(CLI_OECD_Total_Plus_Six[-3:]))
    logger2.info('')
    logger2.info('CLI_Usa: \n' + str(CLI_Usa[-3:]))
    logger2.info('')
    logger2.info('CLI_Korea: \n' + str(CLI_Korea[-3:]))
    logger2.info('')
    logger2.info('CLI_China: \n' + str(CLI_China[-3:]))
    logger2.info('')
    logger2.info('CLI_Germany: \n' + str(CLI_Germany[-3:]))

    plt.figure(figsize=(15,6))
    plt.title(f"Leading indicators: CLI (Composite leading indicator)", fontdict={'fontsize':20, 'color':'g'})
    plt.axvspan(datetime(2020,3,3), datetime(2020,3,30), facecolor='gray', edgecolor='gray', alpha=0.3)
    plt.plot(CLI_OECD_Total, label='CLI_OECD_Total')
    plt.plot(CLI_OECD_Total_Plus_Six, label='CLI_OECD_Total_Plus_Six')
    plt.plot(CLI_Usa, label='CLI_Usa')
    plt.plot(CLI_Korea, label='CLI_Korea')
    plt.plot(CLI_China, label='CLI_China')
    plt.plot(CLI_Germany, label='CLI_Germany')
    plt.grid()
    plt.legend()
    plt.savefig(reports_dir + '/global_0100.png')


'''
1.2 M1
'''
def fed_m1():

    fig, ax = plt.subplots(figsize=(18, 6 * 2))
    # logn term view
    usa = fred.get_series(series_id='WM1NS', observation_start=from_date_LT)
    china = fred.get_series(series_id='MYAGM2CNM189N', observation_start=from_date_LT)
    japan = fred.get_series(series_id='MANMM101JPM189S', observation_start=from_date_LT)
    euro = fred.get_series(series_id='MANMM101EZM189S', observation_start=from_date_LT)
    korea = fred.get_series(series_id='MANMM101KRM189S', observation_start=from_date_LT)

    plt.subplot(2, 1, 1)
    plt.title(f"M1 normalized over the countries over long term", fontdict={'fontsize':20, 'color':'g'})
    # Covid-19 Crisis
    plt.axvspan(datetime(2020,3,3), datetime(2020,3,30), facecolor='gray', edgecolor='gray', alpha=0.3)
    plt.axhline(y=2, linestyle='--', color='red', linewidth=1, label='2% Target Rate')
    plt.plot(normalize(usa), label='usa')
    plt.plot(normalize(china), label='china')
    plt.plot(normalize(japan), label='japan')
    plt.plot(normalize(euro), label='euro')
    plt.plot(normalize(korea), label='korea')
    plt.grid()
    plt.legend()

    # mid term view
    usa = fred.get_series(series_id='WM1NS', observation_start=from_date_MT)
    china = fred.get_series(series_id='MYAGM2CNM189N', observation_start=from_date_MT)
    japan = fred.get_series(series_id='MANMM101JPM189S', observation_start=from_date_MT)
    euro = fred.get_series(series_id='MANMM101EZM189S', observation_start=from_date_MT)
    korea = fred.get_series(series_id='MANMM101KRM189S', observation_start=from_date_MT)
    plt.subplot(2, 1, 2)
    plt.title(f"M1 normalized over the countries over mid term", fontdict={'fontsize':20, 'color':'g'})
    # Covid-19 Crisis
    plt.axvspan(datetime(2020,3,3), datetime(2020,3,30), facecolor='gray', edgecolor='gray', alpha=0.3)

    plt.axhline(y=2, linestyle='--', color='red', linewidth=1, label='2% Target Rate')
    plt.plot(normalize(usa), label='usa')
    plt.plot(normalize(china), label='china')
    plt.plot(normalize(japan), label='japan')
    plt.plot(normalize(euro), label='euro')
    plt.plot(normalize(korea), label='korea')
    plt.grid()
    plt.legend()

    # 이미지 파일로 저장
    plt.tight_layout()  # 서브플롯 간 간격 조절
    plt.savefig(reports_dir + '/global_0120.png')


'''
1.3 CPI (Consumer Price Indices)
'''
def fed_cpi():
    cpi_us = fred.get_series(series_id='CPIAUCSL', observation_start=from_date_MT).pct_change(periods=12)*100
    cpi_japan = fred.get_series(series_id='JPNCPIALLMINMEI', observation_start=from_date_MT).pct_change(periods=12)*100
    cpi_euro = fred.get_series(series_id='CP0000EZ19M086NEST', observation_start=from_date_MT).pct_change(periods=12)*100
    cpi_korea = fred.get_series(series_id='KORCPIALLMINMEI', observation_start=from_date_MT).pct_change(periods=12)*100
    cpi_china = fred.get_series(series_id='CHNCPIALLMINMEI', observation_start=from_date_MT).pct_change(periods=12)*100

    logger2.info('')
    logger2.info('##### CPI (Consumer Price Indices) #####')
    logger2.info('cpi_us: \n' + str(cpi_us[-3:]))
    logger2.info('cpi_japan: \n' + str(cpi_japan[-3:]))
    logger2.info('cpi_euro: \n' + str(cpi_euro[-3:]))  
    logger2.info('cpi_korea: \n' + str(cpi_korea[-3:]))
    logger2.info('cpi_china: \n' + str(cpi_china[-3:]))

    plt.figure(figsize=(18,4))
    plt.title(f"Consumer Price Indices for Countries", fontdict={'fontsize':20, 'color':'g'})
    plt.grid()
    plt.axvspan(datetime(2020,3,3), datetime(2020,3,30), facecolor='gray', edgecolor='gray', alpha=0.3)# Covid-19 Crisis
    plt.plot(cpi_us, label='CPI for U.S. City')
    plt.plot(cpi_japan, label='CPI for Japan')
    plt.plot(cpi_euro, label='CPI for Euro 19 countries')
    plt.plot(cpi_korea, label='CPI for South Korea')
    plt.plot(cpi_china, label='CPI for China')
    plt.legend()
    plt.savefig(reports_dir + '/global_0130.png')





'''
2. Market Area
2.1 Sovereign CDS
'''
def sovereign_cds():
    page = requests.get("https://www.worldgovernmentbonds.com/sovereign-cds/")
    soup = bs(page.text, "html.parser")

    # 제거하려는 태그 선택
    tag_to_remove = soup.find('tfoot')
    # 태그 제거
    tag_to_remove.decompose()

    tables = soup.find_all('table')
    # 멀티 헤더의 첫번째 헤더 제거
    cdses = pd.read_html(str(tables))[0]
    column_to_remove = cdses.columns[0]
    cdses.drop(columns=column_to_remove, inplace=True)
    cdses.columns = cdses.columns.droplevel(0)

    logger2.info('')
    logger2.info('##### Sovereign CDS (Credit Default Swap) #####')
    logger2.info('cdses: \n' + str(cdses))



'''
3. Business Area
3.1 Containerized Freight Index
- 해상운임지수: 경기가 다시 활성화 되는지 여부 모니터링 (20220906)
'''

def container_Freight_index():
    # CCFI (China Containerized Freight Index)
    # 중국컨테이너운임지수는 중국 교통부가 주관하고 상하이 항운교역소가 집계하는 중국발 컨테이너운임지수로 1998년 4월 13일 처음 공시되었다. 
    # 세계컨테이너시황을 객관적으로 반영한 지수이자 중국 해운시황을 나타내는 주요 지수로 평가받고 있다.
    # 1998년 1월 1일을 1,000으로 산정하며 중국의 항구를 기준으로 11개의 주요 루트별 운임을 산정하며, 16개 선사의 운임정보를 기준으로 
    # 매주 금요일에 발표를 하고 있다.
    page = requests.get("https://www.kcla.kr/web/inc/html/4-1_2.asp")
    soup = bs(page.text, "html.parser")
    table = soup.find_all('table')
    df_ccfi = pd.read_html(str(table))[0]

    df_ccfi = df_ccfi.T
    df_ccfi.drop([0], inplace=True)
    df_ccfi[1] = df_ccfi[1].astype('float')
    df_ccfi.rename(columns={0:'Date', 1:'Idx'}, inplace=True)
    df_ccfi['Date']= pd.to_datetime(df_ccfi['Date'])
    df_ccfi.set_index('Date', inplace=True)

    # SCFI (Shanghai Containerized Freight Index)
    # 상하이컨테이너 운임지수는 상하이거래소(Shanghai Shipping Exchange: SSE)에서 2005년 12월 7일부터 상하이 수출컨테이너 운송시장의 
    # 15개 항로의 스팟(spot) 운임을 반영한 운임지수이다. 기존에는 정기용선운임을 기준으로 하였으나 2009년 10월 16일부터는 20ft 컨테이너(TEU)당 
    # 미달러(USD)의 컨테이너 해상화물운임에 기초하여 산정하고 있다.
    # 운송조건은 CY-CY조건이며 컨테이너의 타입과 화물의 상세는 General Dry Cargo Container로 한정짓고 있고, 개별항로의 운임율은 각 항로의 
    # 모든 운임율의 산술평균이며 해상운송에 기인한 할증 수수료가 포함되어 있다. 운임정보는 정기선 선사와 포워더를 포함한 CCFI의 패널리스트들에게 
    # 제공받고 있다.
    page = requests.get("https://www.kcla.kr/web/inc/html/4-1_3.asp")
    soup = bs(page.text, "html.parser")
    table = soup.find_all('table')
    df_scfi = pd.read_html(str(table))[0]

    df_scfi = df_scfi.T
    df_scfi.drop([0], inplace=True)
    df_scfi[1] = df_scfi[1].astype('float')
    df_scfi.rename(columns={0:'Date', 1:'Idx'}, inplace=True)
    df_scfi['Date']= pd.to_datetime(df_scfi['Date'])
    df_scfi.set_index('Date', inplace=True)

    # HRCI (Howe Robinson Container Index)
    # 영국의 대표적인 해운컨설팅 및 브로커社인 Howe Robinson社가 발표하는 컨테이너 지수로서 선박을 하루 용선하는 데 소요되는 비용에 대한
    # 컨테이너 시장 용선요율을 나타내고 있다. 이 회사는 1883년 설립되었으며 컨테이너선과 벌크선에 대한 세계에서 가장 크고 독립적인 중개회사 중 
    # 하나로 1997년 1월 1일을 1,000으로 놓고 매주 발표하고 있다.
    page = requests.get("https://www.kcla.kr/web/inc/html/4-1_4.asp")
    soup = bs(page.text, "html.parser")
    table = soup.find_all('table')
    df_hrci = pd.read_html(str(table))[0]

    df_hrci = df_hrci.T
    df_hrci.drop([0], inplace=True)
    df_hrci[1] = df_hrci[1].astype('float')
    df_hrci.rename(columns={0:'Date', 1:'Idx'}, inplace=True)
    df_hrci['Date']= pd.to_datetime(df_hrci['Date'])
    df_hrci.set_index('Date', inplace=True)


    # BDI (Baltic Dry Index)
    # 발틱운임지수는 발틱해운거래소에서 1999년 11월 1일부터 사용되었으며 1985년부터 건화물(dry cargo)의 운임지수로 사용이 되어온 
    # BFI(Baltic Freight Index)를 대체한 종합운임지수로 1985년 1월 4일을 1,000으로 산정하여 선박의 형태에 따라 발표하고 있다.
    # 선형에 따라 Baltic Capesize Index(BCI), Baltic Panamax Index(BPI), Baltic Supramax Index(BSI), 
    # Baltic Handysize Index(BHSI) 등으로 구성되어 있으며, BDI는 이러한 선형별 정기용선의 4가지 지수를 동일한 가중으로 평균을 산출한 다음 
    # BDI factor를 곱하여 산출하고 있다.
    page = requests.get("https://www.kcla.kr/web/inc/html/4-1_5.asp")
    soup = bs(page.text, "html.parser")
    table = soup.find_all('table')
    df_bdi = pd.read_html(str(table))[0]

    df_bdi = df_bdi.T
    df_bdi.drop([0], inplace=True)
    df_bdi[1] = df_bdi[1].astype('float')
    df_bdi.rename(columns={0:'Date', 1:'Idx'}, inplace=True)
    df_bdi['Date']= pd.to_datetime(df_bdi['Date'])
    df_bdi.set_index('Date', inplace=True)

    plt.figure(figsize=(15,5))
    plt.title(f"Containerized Freight Index", fontdict={'fontsize':20, 'color':'g'})
    plt.grid()
    plt.plot(df_ccfi, label='China') 
    plt.plot(df_scfi, label='Shanghai') 
    plt.plot(df_hrci, label='Howe Robinson') 
    plt.plot(df_bdi, label='Baltic Dry') 
    plt.legend()
    plt.savefig(reports_dir + '/global_0310.png')


# 3.2 Max Draw Down
def daily_returns(prices):
    res = (prices/prices.shift(1) - 1.0)[1:]
    res.columns = ['return']
    return res

def cumulative_returns(returns):
    res = (returns + 1.0).cumprod()
    res.columns = ['cumulative return']
    return res

def max_drawdown(cum_returns):
    max_returns = np.fmax.accumulate(cum_returns)
    res = cum_returns / max_returns - 1
    res.columns = ['max drawdown']
    return res

def max_drawdown_strategy(country:str, tickers:list, assets:list):
    threshold_value = -0.3
    plt.figure(figsize=(16,4*len(tickers)))
    for i, tick in enumerate(tickers):
        if tick == '':
            continue
        ticker = yf.Ticker(tick)
        prices = ticker.history(period='12y')['Close'] # 12: life cycle
        dret = daily_returns(prices)
        cret = cumulative_returns(dret)
        ddown = max_drawdown(cret)
        ddown[ddown.values < -0.3]

        plt.subplot(len(tickers), 1, i + 1)
        plt.grid()
        if assets[i] == 'stock':
            color = 'royalblue'
        elif assets[i] == 'bond':
            color = 'green'
        elif assets[i] == 'commodity':
            color = 'gray'
        else:
            color = 'orange'
        plt.bar(ddown.index, ddown, color=color)
        plt.title(ticker)
        plt.axhline(y=threshold_value, color='red', linestyle='--', label='Threshold')
        plt.xlabel('Date')
        plt.ylabel('Draw Down %')

    plt.tight_layout()  # 서브플롯 간 간격 조절
    plt.savefig(reports_dir + f'/global_0320_{country}.png')



'''
4. calculate trend
- country growth, market growth, business growth
1) cal_trend
    - cal_country_growth
    . get_GDP_rate
    . get_inflation
    . get_export
    - cal_market_growth
    . M2(증가량 배수)
    . 자산시장별 growth
    - cal_busi_growth
    . ETF growth
2) Status decision
    - Status_Now, Status_6m, Status_12m
'''
class CalcuTrend():

    def __init__(self):
        database = 'Economics.db'
        db_file = 'database/' + database
        conn, engine = create_connection(db_file)
        self.conn = conn
        self.engine = engine


    def get_GDP_rate(self, conn, country_sign): # rate 로 측정되는 값은 'actual' 값을 기준으로 가감을 하고,             

        idx = 0
        if country_sign in ['US', 'JP']:
            gdp = pd.read_sql_query(f"SELECT * FROM Calendars WHERE event like 'GDP Growth Rate QoQ%'  AND country = '{country_sign}' \
                ORDER BY date DESC LIMIT 2", conn)
            # print(gdp)
        elif country_sign in ['KR', 'CN', 'DE', 'IN', 'SG', 'BR', 'IN']:
            gdp = pd.read_sql_query(f"SELECT * FROM Calendars WHERE event like '%GDP Growth Rate YoY%'  AND country = '{country_sign}' \
                                    ORDER BY date DESC LIMIT 2", conn)
            # print(gdp)
        else:
            logger.error(f' >>> Country Sign is not found: {country_sign}' )

        
        if np.isnan(gdp['actual'][0]):
            idx = 1

        # 보정 1: 지난 값대비 변화율로 가감점
        mul_1 = (gdp['change'][idx] / 100)
        
        # 보정 2: 예측치 대비 실측치로 가감점
        if np.isnan(gdp['estimate'][idx]):  # gdp['estimate'][idx] is nan 인 경우 초기괎 0 셋팅
            mul_2 = 0
        else:
            mul_2 = (gdp['actual'][idx] - gdp['estimate'][idx]) / 100  
        
        logger2.debug('mul_1: ' + str(mul_1))
        logger2.debug('mul_2: ' + str(mul_2))
        
        result = gdp['actual'][idx] + mul_1 + mul_2
        
        return result
    
    
    def get_inflation(self, conn, country_sign): # inflation mom : price 로 측정되는 값은 'change' 값을 기준으로 가감함.
        inflation = pd.read_sql_query(f"SELECT * FROM Calendars WHERE event like 'Inflation Rate YoY%' \
                                      AND country = '{country_sign}'\
                                      ORDER BY date DESC LIMIT 2", conn)

        # print(inflation)

        try:
            if np.isnan(inflation['estimate'][0]) or np.isnan(inflation['actual'][0]):
                idx = 1
            else:
                idx = 0
        except:
            result = inflation['actual'][0]
            return result
        
        # 보정 1: 지난 월단위값 대비 변화율
        mul_1 = (inflation['change'][idx] / 100)
        
        # 보정 2: 예측치 대비 실측치로 +- 20% 보정
        mul_2 = (inflation['actual'][idx] - inflation['estimate'][idx]) / 100
    
        logger2.debug('mul_1: ' + str(mul_1))
        logger2.debug('mul_2: ' + str(mul_2))    

        result = inflation['actual'][idx] + mul_1 + mul_2
        
        return result
    
    
    def get_export(self, conn, country_sign): # export yoy  : price 로 측정되는 값은 'change' 값을 기준으로 가감함.
        if country_sign in ['US']:            
            export = pd.read_sql_query(f"SELECT * FROM Calendars WHERE country = '{country_sign}' \
                AND event like 'Export Prices YoY%'  ORDER BY date DESC LIMIT 2", conn)
        elif country_sign in ['KR', 'JP', 'CN', 'SG',]:   # SG ..... ??? 넘어가네... 
            export = pd.read_sql_query(f"SELECT * FROM Calendars WHERE country = '{country_sign}' \
                AND event like 'Exports YoY%'  ORDER BY date DESC LIMIT 2", conn)
        elif country_sign in ['DE',]:    
            export = pd.read_sql_query(f"SELECT * FROM Calendars WHERE country = '{country_sign}' \
                AND event like 'Exports MoM%'  ORDER BY date DESC LIMIT 2", conn)  
        elif country_sign in ['IN',]:    
            export = pd.read_sql_query(f"SELECT * FROM Calendars WHERE country = '{country_sign}' \
                AND event like 'Exports%'  ORDER BY date DESC LIMIT 2", conn)              
        else:
            logger2.info(f'Country Sign is not found: {country_sign}')

        # print(export)
        
        try:
            if np.isnan(export['estimate'][0]) or np.isnan(export['actual'][0]):
                idx = 1
            else:
                idx = 0
                
            # 보정 1: 지난 월단위값 대비 변화율
            mul_1 = (export['change'][idx] / 100)

            # 보정 2: 예측치 대비 실측치로 보정
            mul_2 = (export['actual'][idx] - export['estimate'][idx]) / 100                
        except:
            mul_1 = 0
            mul_2 = 0
            idx = 0
            pass
        
        logger2.debug('mul_1: ' + str(mul_1))
        logger2.debug('mul_2: ' + str(mul_2))           
        
        result = export['actual'][idx] + mul_1 + mul_2
        
        return result

    ########################################################################### 
    '''
    I. 국가단위의 현재와 미래의 성장율을 도출하는 루틴으로,
        1. 현재: realGDP실적 성장율(+- 예측치대비)
        1.1 fmp.calendar.com 에서 추출한 realGDP (= nominalGDP - Inflation) 로 계산한 값
            - 잠재성장률대비 real GDP YoY, (include KR.Export) = nominal GDP - Inflation
        1.2 macrovar.com 에서 추출한 realGDP

        2. 미래: (LEI 성장율 예측 (각종 기관들의 값 평균치 활용, IMF, OECD...) + ML perdict) / 2
        2.1 IMF 전망치: 
    '''        
    def cal_country_growth(self, conn, country_sign:str, month_term:int=0):
       
        gdp = self.get_GDP_rate(conn, country_sign)
        inflation = self.get_inflation(conn, country_sign)
        if country_sign in ['EU', 'SG']:
            export = 1
        else:
            export = self.get_export(conn, country_sign)

        logger2.debug(f' {country_sign} gdp: {round(gdp,2)} %')
        logger2.debug(f' {country_sign}  inflation: {round(inflation,2)} %')
        logger2.debug(f' {country_sign}  export: {round(export,2)} %')
        
        realGDP_rate = gdp - inflation*0.3 
        if country_sign in ['KR', 'JP', 'CN']: # 수출 주도형 국가: 한국, 일본, 중국
            result = gdp - inflation*0.3 + export*0.3 # GDP 대비 inflation 비중이 30% 가정, GDP 대비 무역비중이 30% 가정 
        else:
            result = gdp - inflation*0.3 + export*0.03

        return result
    
    ###########################################################################    
    '''
    II. business: M2(증가량 배수) 와 자산시장별 변화율 비교
        - 주식, 채권, 원자재, 부동산, 현금
        - US: S&P500/Nasdaq, 국채3년/10년/30년 수익률, GOLD/OIL, LITZ, DOLLOAR
        - KR: KOSPI/KOSDAQ, 국채3년/10년/30년 수익률, 금현물, LITZ, WON
        - JP: NIKKEI, x, GOLD, LITZ, YEN
        - EU: EUROXXX, BONDS, x, x, x
        - CN: SHANHAE/SIMCHUN, x, x, x, x
        - IN: NIFTY, x, x, x, x
    '''
    def cal_market_growth(self, conn, country_sign1:str, market_name:str, month_term:int): 
        # stock, bond, commidity, currency 별 성장률을 계산하되, bond 는 채권수익률로 계산하는 것임. 채권가격은 보여주는 단에서만.
    
        try:
            # 2.1 M2
            m2 = pd.read_sql_query(f"SELECT * FROM Indicators WHERE Indicator like '%M2%' AND \
            Country = '{country_sign1}' ORDER BY date DESC LIMIT 1", conn)
            m2_growth = m2['YOY'][0]
            if m2['Trend'][0] == 'UP':
                m2_growth = m2_growth * 1.1
            else:
                m2_growth = m2_growth * 0.9
            if m2['Slope'][0] == 'UP':
                m2_growth = m2_growth * 1.05
            else:
                m2_growth = m2_growth * 0.95
        except Exception as e:
            logger.error(' >>> Exception: {}'.format(e))  
            
        #2.2 Assets
        if country_sign1 == 'US':
            if market_name == 'stock':
                ticker = '^SPX'
                ticker = yf.Ticker(ticker)
                sp500 = ticker.history(period='3mo')['Close']                
                sp500 = sp500.dropna()
                growth_1 = (sp500[-1] - sp500[0]) / sp500[0]
                
                ticker = '^IXIC'
                ticker = yf.Ticker(ticker)
                nasdaq = ticker.history(period='3mo')['Close']
                nasdaq = nasdaq.dropna()
                growth_2 = (nasdaq[-1] - nasdaq[0]) / nasdaq[0]  # 최근 것에서 예전 것을 빼고, 이것을 예전 것으로 나누면 성장률

                asset_growth = (growth_1 + growth_2) / 2 * 100
                
            elif market_name == 'bond':  # 10년물 시장금리 기준으로 주식시장과 역의 상관관계로 가정함.
                sleep(0.1)                
                y10 = fred.get_series(series_id='DGS10', observation_start=from_date_ST)
                asset_growth = (y10[0] - y10[-1]) / y10[-1]  # 예전 것에서 최근 것을 뺴고, 이것을 최근 것으로 나누면 채권가격 성장률

                
            elif market_name == 'commodity':
                asset_growth = get_copper()  # 원자재 특히 구리는 미국원자재시장으로 기준으로 한다고 가정
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / currency[-1]
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset')        
            
        elif country_sign1 == 'KR':
            if market_name == 'stock':
                ticker = '^KS200'
                ticker = yf.Ticker(ticker)
                kospi = ticker.history(period='12mo')['Close']                
                kospi = kospi.dropna()
                growth_1 = (kospi[-1] - kospi[0]) / kospi[0]
                
                ticker = '^KQ100'
                ticker = yf.Ticker(ticker)
                kosdaq = ticker.history(period='12mo')['Close']
                kosdaq = kosdaq.dropna()
                growth_2 = (kosdaq[-1] - kosdaq[0]) / kosdaq[0]

                asset_growth = (growth_1 + growth_2) / 2 * 100
                
            elif market_name == 'bond':
                y10 = wgbonds_10y('KR')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계
            elif market_name == 'commodity':
                asset_growth = get_copper()
            elif market_name == 'currency':
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency)  # 역의 상관관계
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset')        

        elif country_sign1 == 'JP':
            if market_name == 'stock':
                ticker = '^N225'
                ticker = yf.Ticker(ticker)
                nikkei = ticker.history(period='3mo')['Close']                
                nikkei = nikkei.dropna()
                growth_1 = (nikkei[-1] - nikkei[0]) / nikkei[0]
                asset_growth = (growth_1) * 100
            elif market_name == 'bond':
                y10 = wgbonds_10y('JP')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계            
            elif market_name == 'commodity':
                asset_growth = get_copper()
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset')        

        elif country_sign1 == 'CN':
            if market_name == 'stock':
                ticker = '000001.SS'
                ticker = yf.Ticker(ticker)
                shanghai = ticker.history(period='3mo')['Close']                
                shanghai = shanghai.dropna()
                if not shanghai.empty:  # shanghai가 비어 있지 않은 경우에만 계산 수행
                    growth_1 = (shanghai.iloc[-1] - shanghai.iloc[0]) / shanghai.iloc[0]
                    # logger2.info(f"shanghai Growth rate:, {growth_1:.2f}")
                else:
                    logger2.info("global_.py Error: No data available for the given period.")
                # growth_1 = (shanghai[-1] - shanghai[0]) / shanghai[0]
                    
                ticker = '399001.SZ'
                ticker = yf.Ticker(ticker)
                Shenzhen = ticker.history(period='3mo')['Close']
                Shenzhen = Shenzhen.dropna()
                if not Shenzhen.empty:
                    growth_2 = (Shenzhen.iloc[-1] - Shenzhen.iloc[0]) / Shenzhen.iloc[0]
                    # logger2.info(f"Shenzhen Growth rate: {growth_1:.2f}")
                else:
                    logger2.info("global_.py Error: No data available for the given period.")                    
                # growth_2 = (Shenzhen[-1] - Shenzhen[0]) / Shenzhen[0]

                asset_growth = (growth_1 + growth_2) / 2 * 100

            elif market_name == 'bond':
                y10 = wgbonds_10y('CN')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계
              
            elif market_name == 'commodity':
                asset_growth = get_copper()
                # print(asset_growth)
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset')       

        elif country_sign1 == 'DE':
            if market_name == 'stock':
                ticker = '^GDAXI'
                ticker = yf.Ticker(ticker)
                dax = ticker.history(period='3mo')['Close']                
                dax = dax.dropna()
                growth_1 = (dax[-1] - dax[0]) / dax[0]

                asset_growth = (growth_1) * 100
                
            elif market_name == 'bond':  # yield
                y10 = wgbonds_10y('DE')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계
                
            elif market_name == 'commodity':
                asset_growth = get_copper()
                # print(asset_growth)
            elif market_name == 'currency':
                currency = fred.get_series(series_id='EXUSEU', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset') 

        elif country_sign1 == 'EU':
            if market_name == 'stock':
                ticker = '^STOXX50E'
                ticker = yf.Ticker(ticker)
                eu = ticker.history(period='3mo')['Close']                
                eu = eu.dropna()
                growth_1 = (eu[-1] - eu[0]) / eu[0]

                asset_growth = (growth_1) * 100
                
            elif market_name == 'bond':
                sleep(0.1)
                y10 = fred.get_series(series_id='IRLTLT01EZM156N', observation_start=from_date_ST)
                asset_growth = (y10[0] - y10[-1]) / len(y10) 
                # print(asset_growth) 
                
            elif market_name == 'commodity':
                asset_growth = get_copper()
                # print(asset_growth)
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='EXUSEU', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset') 

        elif country_sign1 == 'SG':
            if market_name == 'stock':
                ticker = '^STI'
                ticker = yf.Ticker(ticker)
                ses = ticker.history(period='3mo')['Close']                
                ses = ses.dropna()
                growth_1 = (ses[-1] - ses[0]) / ses[0]

                asset_growth = (growth_1) * 100
                
            elif market_name == 'bond':
                y10 = wgbonds_10y('SG')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계
                
            elif market_name == 'commodity':
                asset_growth = get_copper()
                # print(asset_growth)
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset') 

        elif country_sign1 == 'BR':  # Brazil
            if market_name == 'stock':
                ticker = '^FTSE'
                ticker = yf.Ticker(ticker)
                ftse = ticker.history(period='3mo')['Close']                
                ftse = ftse.dropna()
                growth_1 = (ftse[-1] - ftse[0]) / ftse[0]

                asset_growth = (growth_1) * 100
                
            elif market_name == 'bond':
                y10 = wgbonds_10y('BR')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계
                
            elif market_name == 'commodity':
                asset_growth = get_copper()
                # print(asset_growth)
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset') 

        elif country_sign1 == 'IN':
            if market_name == 'stock':
                ticker = '^NSEI'
                ticker = yf.Ticker(ticker)
                nifty50 = ticker.history(period='3mo')['Close']                
                nifty50 = nifty50.dropna()
                growth_1 = (nifty50[-1] - nifty50[0]) / nifty50[0]

                asset_growth = (growth_1) * 100
                
            elif market_name == 'bond':
                y10 = wgbonds_10y('IN')
                asset_growth = (y10.iloc[-1,1] - y10.iloc[0,1]) / y10.iloc[0,1] # 역의 상관관계
                
            elif market_name == 'commodity':
                asset_growth = get_copper()
                # print(asset_growth)
            elif market_name == 'currency':
                sleep(0.1)                
                currency = fred.get_series(series_id='RTWEXBGS', observation_start=from_date_MT)
                asset_growth = (currency[0] - currency[-1]) / len(currency) 
                # print(asset_growth)
            else:
                logger.error(f' >>> Error: {country_sign1}  {market_name} Not found Asset')                 

        else:
            logger.error(f' >>> Error: {country_sign1}: Country sign is not found.')
         
        result = (m2_growth * 0.6) + (asset_growth * 0.4)
        
        logger2.debug(f' {country_sign1} Market Growth: {round(result, 2)} %')           
        logger2.debug(f' {country_sign1} M2 Growth: {round(m2_growth,2)} %')
        logger2.debug(f' {country_sign1} Asset Growth: {round(asset_growth,2)} %')

        return result


    ###########################################################################    
    '''
    III. business: 각 섹터별 ETF growth 
        - 특별히 업종을 주도하는 섹터만 집중 투자하는 경우만 활용
        - US: spy/qqq, shy/tlt/tmf, gld/bci, o/vnq, dollar
        - kr: kodex200/tiger200it, kodex국고채3년/kosef국고채10년/kbstar kis국고채30년enhanced, 금현물, 맥쿼리인프라, 원
        - jp: 노무라닛케이225etf, yen
        - eu:
        - cn:  
        month_term: 0, 3, 12, 36 개월후 예측치 산정
    '''
    def cal_busi_growth(self, conn, ticker:str, month_term:int=0):
        df = pd.DataFrame() # Empty DataFrame
        try:
            df = pd.read_csv(f"batch/reports/data/{ticker}.csv", sep=",")
            # VWAP requires the DataFrame index to be a DatetimeIndex.
            # Replace "datetime" with the appropriate column from your DataFrame
            df.set_index(pd.DatetimeIndex(df["date"]), inplace=True)
            df = df[df['date'] >= from_date_ST2]            
        except:
            ticker = yf.Ticker(ticker)
            df = ticker.history(start=from_date_MT2)
            df.set_index(pd.DatetimeIndex(df.index), inplace=True)
            df = df[df.index >= from_date_ST2]                

        # Calculate Returns and append to the df DataFrame
        df.ta.percent_return(cumulative=True, append=True)
        result = df['CUMPCTRET_1'][-1] * 100

        return result


    ###################################
    '''
    IV. 국가/시장/사업단위의 성장률을 근거로 트랜드, 사이클상 현위치, 그리고 현재/6개월/12개월 전망을 도출하는 마지막 단계.
        - 입력받은 연구기관, 국가/시장/사업단위 성장률 정보와 Alpha 테이블 history 정보를 기반으로 
          사이클상의 어느 위치에 와 있으며, 6개월, 12개월 전망은 어떻게 될 것인지 판단하는 루틴
        - OECD 는 2023년말 기준으로 2개년치 2024, 2025년 전망치 제공
        - IMF 는 2023년말 기준으로 5개년치 2024, 2025, 2026, 2027, 2028년 전망치 제공
        - World Bank
        - JP Morgan..
    '''

    # ##############################################################
    # IV-1. OECD LEI 현재월값은 향후 6개월이내 단기 경제지표(반기단위 발표) 발표. 
    # - 수정한 월 + 6개월 전망으로 가정함. 중기 전망으로는 사용할수 없음
    # - 향후 전망시 6개월 이전까지는 OECD 예상치로, 6개월 이후는 IMF 전망치로 산정
    # ##############################################################
    def get_country_growth_fore_byOECD(self, conn, country_sign3:str, month_term:int=6):  # country_sign3: United States
        
        def add_month(to_date2:str, term_month:int):
            target_date = pd.to_datetime(to_date2)
            term_month = relativedelta(weeks=term_month*4)
            target_date2 = (target_date + term_month).date()
            return target_date2

        target_date2 = add_month(to_date, month_term)
        logger2.debug(f" target date: {target_date2}")
        
        # Gross domestic product, market prices, deflator, growth
        if country_sign3 == "China (People's Republic of)":
            val = pd.read_sql_query(f"SELECT Value FROM OECD WHERE Country like 'China%' \
                                    AND Variable = 'Gross domestic product, market prices, deflator, growth' \
                                    AND Time = '{target_date2.year}'", conn)
        else:
            val = pd.read_sql_query(f"SELECT Value FROM OECD WHERE Country = '{country_sign3}' \
                                    AND Variable = 'Gross domestic product, market prices, deflator, growth' \
                                    AND Time = '{target_date2.year}'", conn)
        result = float(val.values)
        logger2.debug(f' {country_sign3} GDP Growth: {result} %')         
                                                        
        return result
  
    # ##############################################################
    # IV-2. IMF LEI 를 중기 전망(month_term > 6)으로 사용함.
    # - 향후 전망시 6개월 이전까지는 OECD 예상치로, 6개월 이후는 IMF 전망치로 산정
    # ##############################################################    
    def get_country_growth_fore_byIMF(self, conn, country_sign2:str, month_term:int=6):
        
        def add_month(to_date2:str, term_month:int):
            target_date = pd.to_datetime(to_date2)
            term_month = relativedelta(weeks=term_month*4)
            target_date2 = (target_date + term_month).date()
            return target_date2
        
        target_date2 = add_month(to_date, month_term)
        logger2.debug(f" target date: {target_date2}")
        target_year = target_date2.year
        lei_gdp_rate = pd.read_sql_query(f"SELECT * FROM IMF WHERE  ISO = '{country_sign2}' AND \"WEO Subject Code\" = 'NGDP_RPCH'", conn)
        lei_inflation_rate = pd.read_sql_query(f"SELECT * FROM IMF WHERE  ISO = '{country_sign2}' AND \"WEO Subject Code\" = 'PCPIPCH'", conn)
        lei_export_rate = pd.read_sql_query(f"SELECT * FROM IMF WHERE  ISO = '{country_sign2}' AND \"WEO Subject Code\" = 'TXG_RPCH'", conn)
        logger2.debug(f' {country_sign2} lei_gdp_rate:' + str(lei_gdp_rate))
        logger2.debug(f' {country_sign2} lei_inflation_rate:' + str(lei_inflation_rate))
        logger2.debug(f' {country_sign2} lei_export_rate:' + str(lei_export_rate))
 
        gdp = lei_gdp_rate[str(target_year)]
        inflation = lei_inflation_rate[str(target_year)]
        export = lei_export_rate[str(target_year)]
        logger2.debug(f'target year: ' + str(target_year))
        logger2.debug(f'gdp: ' + str(gdp))
        logger2.debug(f'inflation: ' + str(inflation))
        logger2.debug(f'export: ' + str(export))    
        
        realGDP_rate = float(gdp) - float(inflation) * 0.3  # 한국에서 GDP 대비 소비비중????
        if country_sign2 in ['KOR', 'JPN', 'CHN']: # 수출 주도형 국가: 한국, 일본, 중국
            result = realGDP_rate + float(export) * 0.3 # GDP 대비 무역비중이 30% 가정 
        elif country_sign2 in ['USA']:
            result = realGDP_rate + float(export) * 0.05 # GDP 대비 무역비중이 0% 가정 
        else:
            result = realGDP_rate + float(export) * 0.2 # 미국 빼고, 수출형 국가 빼고 그 나머지들.

        return result

    # ##############################################################
    # IV-2. IMF LEI 를 중기 전망(month_term > 6)으로 사용함.
    # - 향후 전망시 6개월 이전까지는 OECD 예상치로, 6개월 이후는 IMF 전망치로 산정
    # ##############################################################    
    def get_country_growth_fore_byWorldBank(self, conn, country_sign3:str, month_term:int=6):
        
        def add_month(to_date2:str, term_month:int):
            target_date = pd.to_datetime(to_date2)
            term_month = relativedelta(weeks=term_month*4)
            target_date2 = (target_date + term_month).date()
            return target_date2

        target_date2 = add_month(to_date, month_term)
        logger2.debug(f" target date: {target_date2}")
        
        # Gross domestic product, market prices, deflator, growth
        if country_sign3 == "United States":
            if target_date2.year == 2024:
                buff = f"SELECT _2024f FROM WorldBank WHERE Category_3 = 'United States'"
            elif target_date2.year == 2025:
                buff = f"SELECT _2025f FROM WorldBank WHERE Category_3 = 'United States'"
            else:
                logger.error(' >>> World Bank Target Date is not valid1.')
        elif country_sign3 == "China (People's Republic of)":
            if target_date2.year == 2024:
                buff = f"SELECT _2024f FROM WorldBank WHERE Category_4 = 'China'"
            elif target_date2.year == 2025:
                buff = f"SELECT _2025f FROM WorldBank WHERE Category_4 = 'China'"
            else:
                logger.error(' >>> World Bank Target Date is not valid2.')
        elif country_sign3 == "Japan":
            if target_date2.year == 2024:
                buff = f"SELECT _2024f FROM WorldBank WHERE Category_3 = 'Japan'"
            elif target_date2.year == 2025:
                buff = f"SELECT _2025f FROM WorldBank WHERE Category_3 = 'Japan'"
            else:
                logger.error(' >>> World Bank Target Date is not valid3.')                    
        elif country_sign3 == "Germany":
            if target_date2.year == 2024:
                buff = f"SELECT _2024f FROM WorldBank WHERE Category_3 = 'Euro area'"
            elif target_date2.year == 2025:
                buff = f"SELECT _2025f FROM WorldBank WHERE Category_3 = 'Euro area'"
            else:
                logger.error(' >>> World Bank Target Date is not valid4.')
        elif country_sign3 == "India":
            if target_date2.year == 2024:
                buff = f"SELECT _2024f FROM WorldBank WHERE Category_4 = 'India 2'"
            elif target_date2.year == 2025:
                buff = f"SELECT _2025f FROM WorldBank WHERE Category_4 = 'India 2'"
            else:
                logger.error(' >>> World Bank Target Date is not valid4.')                    
        else:
            logger.error(' >>> country_sign3 is not found at World Bank.')

        val = pd.read_sql_query(buff, conn)                                    
        result = float(val.values)
        logger2.debug(f' {country_sign3} GDP Growth: {round(result,2)} %')
                                                        
        return result
        
    # ##############################################################    
    # 국가/시장/사업단위의 성장률을 근거로 트랜드, 사이클상 현위치, 그리고 현재/6개월/12개월 전망을 도출하는 마지막 단계.
    # - 입력받은 연구기관, 국가/시장/사업단위 성장률 정보와 Alpha 테이블 history 정보를 기반으로 
    #   사이클상의 어느 위치에 와 있으며, 6개월, 12개월 전망은 어떻게 될 것인지 판단하는 루틴
    # ##############################################################    
    def cal_trend(self, country:str, market:str, business:str, researcher:str, month_term:int):  # country = 'US'. 'KR'...

        ticker = business
        contury_sign1 = country  # country: KR, US, JP...
        country_sign2 = COUNTRIES[contury_sign1][0]['alpha3']  # KOR, USA, JPN....
        country_sign3 = COUNTRIES[contury_sign1][1]['name']  # Republic of Korea, United States, Japan..

        logger2.debug('')    
        logger2.debug(f'##### {contury_sign1} / {market} / {business}')               

        if month_term == 0:
            c_growth = self.cal_country_growth(self.conn, contury_sign1, month_term)
            m_growth = self.cal_market_growth(self.conn, contury_sign1, market, month_term)
            b_growth = self.cal_busi_growth(self.conn, ticker, month_term)
        else:
            if researcher == 'OECD':
                # country_sign2 = COUNTRIES[country][0]['alpha3']
                if contury_sign1 in ['SG']:
                    c_growth = 2.2  # 23년 성장률 값으로 대체, OECD outlook 대상 아님.
                else:
                    c_growth = self.get_country_growth_fore_byOECD(self.conn, country_sign3, month_term)
                m_growth = self.cal_market_growth(self.conn, contury_sign1, market, month_term)
                b_growth = self.cal_busi_growth(self.conn, ticker, month_term)
            elif researcher == 'IMF': # IMF 전망치만 적용
                # country_sign2 = COUNTRIES[contury_sign1][0]['alpha3']
                c_growth = self.get_country_growth_fore_byIMF(self.conn, country_sign2, month_term)
                m_growth = self.cal_market_growth(self.conn, contury_sign1, market, month_term)
                b_growth = self.cal_busi_growth(self.conn, ticker, month_term)
            elif researcher == 'WorldBank':
                if country_sign3  in ['United States', 'Japan', "China (People's Republic of)", 'India']:
                    # country_sign2 = COUNTRIES[country][0]['alpha3']
                    c_growth = self.get_country_growth_fore_byWorldBank(self.conn, country_sign3, month_term)

                else:
                    c_growth = 1 # 결국 이전값 대비 변화율이 필요하므로 괜찮음.

                m_growth = self.cal_market_growth(self.conn, contury_sign1, market, month_term)
                b_growth = self.cal_busi_growth(self.conn, ticker, month_term)                
            else:
                logger.error('>>> researcher is not found.')
           
    
        logger2.debug(f' {contury_sign1} Country Growth: {round(c_growth,2)} %')           
        logger2.debug(f' {contury_sign1} Market Growth: {round(m_growth,2)} %')
        logger2.debug(f' {contury_sign1} Business Growth: {round(b_growth,2)} %')    

        # m_growth 0.3 의 비중으로, b_growth 0.3*0.7= 0.21 의 비중으로 경험치 산정하였음, 추후 알고리즘으로 다이나믹 상수로 만들 예정. 20240220
        # m_growth 에서 bond 마켓은 수익률로 음의 값이 긍정이므로 이에 대한 역보정을 해야 TLT,TMF 의 가격이 긍정으로 작용토록 함.
        if market == 'bond':
            trend = c_growth + (m_growth*0.3)*-1 + b_growth*0.3*0.7
        else:
            trend = c_growth + m_growth*0.3 + b_growth*0.3*0.7
        
        return trend, c_growth, m_growth, b_growth

# ============================= Calss End ============================= #



'''
글로벌 국가들의 트랜드분석 함수
'''
def calculate_trend():

    obj_trend = CalcuTrend()
    # researcher = 'WorldBank'
    # month_terms = [0, 3, 6, 12, 18, 24]  # 현재와 n 개월 전망치 값
    df_alpha = pd.DataFrame(columns=['Country', 'Market', 'Busi', 'Researcher', 'Date', 'Country_Growth', 'Market_Growth',\
                                'Busi_Growth', 'Trend', 'Trend_3mo', 'Trend_6mo', 'Trend_12mo', 'Trend_18mo', 'Trend_24mo'],
                                index=['Country', 'Market', 'Busi', 'Researcher', 'Date'])

    for nation, assets in WATCH_TICKERS2.items():  # 국가별

        if nation in ['EU']:  # 몇 가지 정보가 존재하지 않아 제외
            continue
            
        for asset_grp in assets:  # 국가별 / 자산별 /

            for asset, tickers in asset_grp.items():  # 리스트에서 키와 아이템 분리용 => 딕셔너리 of 리스트 형태 자료구조론임.

                for ticker in tickers:  # 국가별 / 자산별 / ETF별

                    if ticker == '':
                        continue                    

                    for researcher in RESEARCHERS:  # 국가별 / 자산별 / ETF별 / 연구기관별

                        _trend_0 = 0
                        _trend_3 = 0
                        _trend_6 = 0
                        _trend_12 = 0
                        _trend_18 = 0
                        _trend_24 = 0

                        for month_term in month_terms:  # 국가별 / 자산별 / ETF별 / 연구기관별 / 전망월별 (현재부터 24개월후까지 6개월간)

                            # trend, c_growth, m_growth, b_growth = obj_trend.cal_trend(nation, asset, ticker, researcher, month_term)
                            trend, c_growth, m_growth, b_growth = obj_trend.cal_trend(nation, asset, ticker, researcher, 0)
                            logger2.debug(f'##### {researcher} total Trend {nation}/{asset}/{ticker} : {round(trend,2)} %')

                            if month_term == 0:
                                _trend_0 = round(trend, 3)
                            elif month_term == 3:
                                _trend_3 = round(trend, 3)                            
                            elif month_term == 6:
                                _trend_6 = round(trend, 3)
                            elif month_term == 12:
                                _trend_12 = round(trend, 3)
                            elif month_term == 18:
                                _trend_18 = round(trend, 3)
                            else:
                                _trend_24 = round(trend, 3)
                                
                        buffer = pd.DataFrame()
                        buffer['Country'] = [nation]
                        buffer['Market'] = [asset]
                        buffer['Busi'] = [ticker]
                        buffer['Researcher'] = [researcher]
                        buffer['Date'] = [pd.to_datetime(to_date2).date()]
                        buffer['Country_Growth'] = [round(c_growth, 3)]
                        buffer['Market_Growth'] = [round(m_growth, 3)]
                        buffer['Busi_Growth'] = [round(b_growth, 3)]
                        buffer['Trend'] = [_trend_0]
                        buffer['Trend_3mo'] = [_trend_3]
                        buffer['Trend_6mo'] = [_trend_6]
                        buffer['Trend_12mo'] = [_trend_12]
                        buffer['Trend_18mo'] = [_trend_18]
                        buffer['Trend_24mo'] = [_trend_24]
                        logger2.info(buffer)

                        df_alpha = pd.concat([df_alpha, buffer])
                    
    return df_alpha



'''

'''
def plot_alpha_tickers(conn):

    M_table = 'Alpha'
    M_countries = ['US', 'KR', 'JP', 'CN', 'DE', 'IN', 'SG']


    def add_month(to_date2:str, term_month:int):
        target_date = pd.to_datetime(to_date2)
        term_month = relativedelta(weeks=term_month*4)
        target_date2 = (target_date + term_month).date()
        return target_date2
    
    for month_term in month_terms:
        if month_term == 0:
            col_0 = add_month(to_date_2, month_term)
        elif month_term == 3:
            col_3 = add_month(to_date_2, month_term)
        elif month_term == 6:
            col_6 = add_month(to_date_2, month_term)
        elif month_term == 12:
            col_12 = add_month(to_date_2, month_term)
        elif month_term == 18:
            col_18 = add_month(to_date_2, month_term)                                    
        else:
            col_24 = add_month(to_date_2, month_term)


    for country in M_countries:

            M_query = f"SELECT * from {M_table} WHERE Country = '{country}' GROUP BY Country, Market, Busi, Researcher"

            try:
                df = pd.read_sql_query(M_query, conn)
                df = df.drop(columns=['Date', 'Country_Growth', 'Market_Growth', 'Busi_Growth'])
                df = df.sort_values(['Country', 'Market', 'Busi', 'Researcher'], ascending=True).reset_index(drop=True)
                df.columns = ['Country', 'Market', 'Busi', 'Researcher', f'{col_0}', f'{col_3}', f'{col_6}', f'{col_12}', f'{col_18}', f'{col_24}']

                # logger2.info(df.head(5))
                melted_df = pd.melt(df, id_vars=['Country', 'Market', 'Busi', 'Researcher'],
                            var_name=f'{col_0}', value_name='Growth') # @@@ var_name=f'{col_0}'... 우연...
            except Exception as e:
                logger.error(' >>> plot_alpha_tickers Exception: {}'.format(e))

            events = df['Busi'].unique()

            # 전체 그림의 크기를 설정
            plt.figure(figsize=(12, 4*len(events)))
            for i, event in enumerate(events):
                result = melted_df[melted_df['Busi'] == event]        
                if result.empty:
                    continue
                if result['Growth'].all() == 0:
                    continue
                result.dropna()
                country = result.iloc[0]['Country']
                market = result.iloc[0]['Market']
                busi = result.iloc[0]['Busi']
                result.columns = ['Country','Market','Busi','Researcher','Time','Growth']
                result = result.sort_values(['Researcher','Time'], ascending=True).reset_index(drop=True)
                result['Time'] = pd.to_datetime(result['Time'])
                plt.subplot(len(events), 1, i + 1)
                for researcher in RESEARCHERS:
                    buf = result[result['Researcher'] == researcher]
                    if researcher == 'OECD':
                        linestyle = '-'
                        color = 'royalblue'
                    elif researcher == 'IMF':
                        linestyle = '-'
                        color = 'green'
                    else:
                        linestyle = '-'
                        color = 'gray'                        
                    plt.plot(buf['Time'], buf['Growth'], label=researcher, linewidth=1.2, color=color, linestyle=linestyle)
                max_val = max(result['Growth'])
                min_val = min(result['Growth'])
                if (max_val > 0) and (min_val < 0):       # 시각효과     
                    plt.axhline(y=0, linestyle='--', color='red', linewidth=1)
                plt.title(f"Alpha Expectation Outlook {country}: {market} / {busi}")
                plt.grid()
                plt.xlabel('Time')
                plt.ylabel('Growth')
                plt.legend()

            plt.tight_layout()  # 서브플롯 간 간격 조절
            plt.savefig(reports_dir + f'/global_0200_{country}.png')    



'''
Main Fuction
'''

if __name__ == "__main__":


    '''
    0. 공통 함수
    '''
    df_alpha = calculate_trend()
    make_alpha(df_alpha)  # Alpha 테이블 구성작업

    '''
    1. Economic Area
    '''
    oecd_cli()
    fed_cli()
    fed_m1()
    fed_cpi()
    sovereign_cds()
    
    '''
    2. Market Area
    2.1 해상운임지수
    '''
    container_Freight_index()    



    '''
    3. Business Area
    3.1 국가별 Maximum drawdown
    '''
    for nation, assets in WATCH_TICKERS.items():  # 국가별
        buf = []  # ticker 들 모두 나열
        buf2 = []  # ticker 들 모두 나열한 것들의 asset 명 나열
        for asset_grp in assets:  # 국가별 / 자산별 /
            for asset, tickers in asset_grp.items():  # 리스트에서 키와 아이템 분리용 => 딕셔너리 of 리스트 형태 자료구조론임.
                buf.append(tickers)
                for tick in tickers:
                    buf2.append(asset)

        tot_tickers = [item for subs in buf for item in subs]
        
        logger2.info(tot_tickers)
        max_drawdown_strategy(nation, tot_tickers, buf2) # max draw down strategy : 바닥에서 분할 매수구간 찾기




    '''
    3.2 국가별 트랜드 분석 with OECD, IMF, WorldBank 기준으로 ticker 별 현재/3/6/12/18/24 개월후 전망분석
    next time: GS, JPM...
    '''
    plot_alpha_tickers(conn)



# Copyright 2023-2025 Jeongmin Kang, jarvisNim @ GitHub
# See LICENSE for details.
'''
Prgram 명: technical Analysis 만들기
Author: jeongmin Kang
Mail: jarvisNim@gmail.com
History
- 20240228  Create
'''
import sys, os, time
import numpy as np
import pandas as pd
import pandas_ta as ta
import matplotlib.pyplot as plt
import yfinance as yf

from scipy import signal
from datetime import date, datetime, timedelta
from .utils import constant as cst
# from .utils.techs_funcs import (
#     find_30days_ago,
# )
from . import base

'''
0. 공통영역 설정
'''
# logging
base.logger.warning(sys.argv[0] + ' :: ' + str(datetime.today()))
base.logger2.info('')
base.logger2.info(sys.argv[0] + ' :: ' + str(datetime.today()))

#####################################
# funtions
#####################################

# 3개월단위로 순차적으로 읽어오는 경우의 시작/종료 일자 셋팅
# to_date_2 = pd.to_datetime(to_date2)
# three_month_days = relativedelta(weeks=12)
# from_date = (to_date_2 - three_month_days).date()
# to_date_2 = to_date_2.date()

# TECH_LIST = ['sma','ema','macd','adx','psar','ichmoku',  # Trend
#            'rsi','stoch','roc','cci','willr','ao','stochrsi','ppo',  # Momentum
#            'obv','pvt','pvi','cmf','vwap','adosc','mfi','kvo','nvi',  # Volume
#            'atr','bbands','donchian','kc','rvi',]  # Volatility

def get_tech_types(tech_type):
    trend_indics = ['sma','ema','macd','adx','psar','ichmoku','obv','vwap','adosc','atr','donchian',
                    'rvi',]
    overBoughtSold_indics = ['rsi','stoch','willr','mfi','pvi',]
    trendReversal_indics = ['roc','ao','nvi','bbands','kc',]
    kReversal_indics = ['kvo','cmf',]

    if tech_type in trend_indics:
        result = 'trend'
    elif tech_type in overBoughtSold_indics:
        result = 'overBoughtSold'
    elif tech_type in trendReversal_indics:
        result = 'trendReversal'
    elif tech_type in kReversal_indics:
        result = 'kReversal'
    else:
        result = None

    return result

# yfinance 를 이용한 ticker 히스토리 데이터 가져오기
def get_tech_hist(ticker, window):

    ticker_obj = yf.Ticker(ticker)
    # History Data
    df_hist = ticker_obj.history(period='36mo') # test: 10mo, real: 36mo
    try:
        df_hist['feature'] = signal.detrend(df_hist['Close'])  # 통계 정확성을 위한 주기성 시그널 제거용
    except ValueError as e: # ValueError: cannot reshape array of size 0 into shape (0,newaxis)
            time.sleep(3.1)
            base.logger2.error(f'{ticker} get_tech_hist Value error 1: {e}')  # get_tech_hist Value error 1: array must not contain infs or NaNs: 채권에서처럼 detrend 를 만들수 없는 오류
            df_hist = ticker_obj.history(period='36mo') # test: 10mo, real: 36mo
    df_hist['mean'] = df_hist['feature'].rolling(window=window).mean()
    df_hist['std'] = df_hist['feature'].rolling(window=window).std()

    return df_hist

# yfinance 를 이용한 ticker statesments 데이터 가져오기
def get_tech_fin(ticker):

    ticker_obj = yf.Ticker(ticker)
    # Financial Data: 이 정보는 finviz 에서 분기단위 + 4개분기 더 제공하여 대체됨. (finviz fobidden 시 활용가능) 20240309
    pnl = ticker_obj.financials  # 손익계산서
    bs = ticker_obj.balancesheet  # 재무재표, 재정상태표
    cf = ticker_obj.cashflow  # 현금흐름표
    # concatenate into one dataframe
    _df_fin = pd.concat([pnl, bs, cf])
    # make dataframe format nicer
    # Swap dates and columns
    df_fin = _df_fin.T
    # reset index (date) into a column
    df_fin = df_fin.reset_index()
    # Rename old index from '' to Date
    df_fin.columns = ['Date', *df_fin.columns[1:]]

    return df_fin


def make_tech_plot(ticker, df, tech_type, TECH_LIST, idx):
    
    _df = df.copy() # 그래프만 그리면 되고, 본 df 에는 영향을 주지않기위해 복사본 만듬.
    # _df = _df.dropna()  # @@@ 그래프 종료일자까지 일부 그래프가 달라지는 현상이 있어 검증을 위해 제거

    h_limit = 70  # 밴드형 상단 기준선
    l_limit = 30  # 밴드형 하단 기준선
    base_limit = 0  # 기준선형 기준선
    permit_range = 0.02
    tech_list = []
    result = 0  # 테크니컬 분석 결과, latest date 에 해당하는 분선결과 값: 2 매수시작, 1매수마무리, 0대기, -2매도시작, -1매도마무리

    for group, techs  in TECH_LIST.items():
        for tech in techs:
            tech_list.append(tech)

    plt.subplot(len(tech_list), 1, idx)

    max_val = max(_df[tech_type])
    min_val = min(_df[tech_type])

    if tech_type in ['sma','ema','macd','adx','ppo','psar','ichmoku','stoch','roc','ao',
                     'obv','pvt','pvi','cmf','vwap','adosc','kvo','nvi','atr','rvi',]:  # 기준선형 그래프로
        
        if tech_type in ['obv','pvt','cmf','vwap','adosc']:  # Volume 베이스 들은 수치가 커서 지표그래프 안보임.
            pass
        else:
            plt.plot(_df.index, _df['Close'], color='royalblue')

        if tech_type == 'adx':
            offset = 25  # offset: DI+ 와 DI- 차이가 25% 이상 또는 이하이면 signal ON
            _df[tech_type] = _df[tech_type] - offset
            plt.plot(_df.index, _df['DMP_14'], color='orange')
            plt.plot(_df.index, _df['DMN_14'], color='orange')
        elif tech_type == 'stoch':
            h_limit = 85
            l_limit = 15            
            plt.plot(_df.index, _df['STOCHk_14_3_3'], color='orange')
            plt.plot(_df.index, _df['STOCHd_14_3_3'], color='orange')
            plt.axhline(y=h_limit, linestyle='--', lw=1.4, color='gray',)
            plt.axhline(y=l_limit, linestyle='--', lw=1.4, color='gray',)            

        plt.bar(_df.index, _df[tech_type], color=['g' if _df[tech_type].iloc[i] > base_limit else 'r' for i in range(len(_df))])

        # 이전에 널리 알려져있는 신호포착 방법
        # _df['pivot'] = np.where((_df[tech_type] > base_limit) & (_df[tech_type].shift(1) <= base_limit), 1, 0)  # Buy signal
        # _df['pivot'] = np.where((_df[tech_type] < base_limit) & (_df[tech_type].shift(1) >= base_limit), -1, _df['pivot']) # Sell signal

        # 분할매수 시작:2, 분할매수 마무리: 1,  분할매도 시작: -2, 분할매도 마무리: -1
        # buttom에서 매수시그널 리포트, 분할매수 시작, 기준선까지가 분할매수 마무리하고,
        # 그 상태로 홀딩하다다가,
        # top 에서 매도시그널 리포트, 분할매도 시작, 기준선까지 분할매도 마무리하는 방식으로 전환: 20240305
        _df['SellAndBuy'] = _df[tech_type].ewm(span=5, adjust=False, min_periods=5).mean()
        try:
            slope_before = base.trend_detector(_df, 'SellAndBuy', base.get_working_day_before(5))
        except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
            slope_before = (df[tech_type] - df[tech_type].shift(3)) / 3
        print('        ', slope_before)

        # Here is Buttom, 분할매수 시작, slope 이 마이너스(-)에서 0 로 전환된 날 리포트 : 2
        _df['pivot'] = np.where((slope_before <= abs(permit_range)) & (_df['SellAndBuy'].shift(1) < _df['SellAndBuy']) & \
                                _df['SellAndBuy'] < 0, 2, 0)
        # 분할매수 끝, slope 이 플러스(+) 인 상태에서, 이동평균값이 기준선 0 위로 전환된 날 리포트 : 1 
        _df['pivot'] = np.where((slope_before > permit_range) & (_df['SellAndBuy'].shift(1) < _df['SellAndBuy']) & \
                                _df['SellAndBuy'] < 0, 1, _df['pivot'])

        # Here is Top, 분할매도 시작, slope 이 플러스(+)에서 0 으로 전환된 날 리포트 : -2
        _df['pivot'] = np.where((slope_before <= abs(permit_range)) & (_df['SellAndBuy'].shift(1) > _df['SellAndBuy']) & \
                                _df['SellAndBuy'] > 0, -2, _df['pivot'])
        # 분할매도 끝, slope 이 마이너스(-)인 상태에서 , 이동평균값이 기준선 0 아래로 전환된 날 리포트: -1
        _df['pivot'] = np.where((slope_before < permit_range*(-1)) & (_df['SellAndBuy'].shift(1) > _df['SellAndBuy']) & \
                                _df['SellAndBuy'] > 0, -1, _df['pivot'])

        _1 = _df[_df['pivot'] == 2] # 분할매수 시작
        _2 = _df[_df['pivot'] == 1] # 분할매수 마무리
        _3 = _df[_df['pivot'] == -2] # 분할매도 시작
        _4 = _df[_df['pivot'] == -1] # 분할매도 마무리

        if not _1.empty:
            print(_1.iloc[-1]['pivot'])
        if not _2.empty:            
            print(_2.iloc[-1]['pivot'])
        if not _3.empty:            
            print(_3.iloc[-1]['pivot'])
        if not _4.empty:            
            print(_4.iloc[-1]['pivot'])

        plt.axhline(y=base_limit, linestyle='--', color='red', linewidth=1)

    else:  # top과 Buttom 구간값을 기준으로 설정하는 밴드형인경우    

        if tech_type == 'cci':  
            h_limit = 100
            l_limit = -100
        elif tech_type == 'willr':
            h_limit = -20
            l_limit = -80

        plt.plot(_df.index, _df['Close'], color='royalblue')

        # 분할매수 시작:2, 분할매수 마무리: 1,  분할매도 시작: -2, 분할매도 마무리: -1
        # buttom에서 매수시그널 리포트, 분할매수 시작, 기준선 하단까지가 분할매수 마무리하고,
        # 그 상태로 홀딩하다다가,
        # top 에서 매도시그널 리포트, 분할매도 시작, 기준선 상단까지 분할매도 마무리하는 방식으로 전환: 20240305
        _df['SellAndBuy'] = _df[tech_type].ewm(span=5, adjust=False, min_periods=5).mean()
        try:
            slope_before = base.trend_detector(_df, 'SellAndBuy', base.get_working_day_before(5))
        except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
            slope_before = (df[tech_type] - df[tech_type].shift(3)) / 3
        print('        ', slope_before)

        if tech_type == 'bbands':  # 
            plt.plot(_df.index, _df['BBU_5_2.0'], color='orange')  # Upper line
            plt.plot(_df.index, _df['BBL_5_2.0'], color='orange')  # Lower line
            # print(_df)
            # 분할매도는 상단기준선 넘으면서 바로 시작 (Not top) 해서 top 에서 매도 마무리
            _df['pivot'] = np.where((_df[tech_type] >= _df['BBU_5_2.0']) & (_df[tech_type].shift(1) <= _df['BBU_5_2.0']).shift(1), -2, 0)
            # 분할매도 마무리
            _df['pivot'] = np.where((_df[tech_type] >= _df['BBU_5_2.0']) & (_df[tech_type].shift(1) >= _df['BBU_5_2.0']).shift(1) & \
                                    (slope_before <= abs(permit_range)), -1, _df['pivot'])
            # 분할매수 시작은 Buttom 에서 시작해서 기준선 하단 통과시 매수 마무리
            _df['pivot'] = np.where((_df[tech_type] <= _df['BBL_5_2.0']) & (slope_before <= abs(permit_range)) & \
                                    (_df[tech_type].shift(1) <= _df['BBL_5_2.0']).shift(1), 2, _df['pivot'])
            # 분할매수 마무리
            _df['pivot'] = np.where((_df[tech_type] <= _df['BBL_5_2.0']) & (_df[tech_type].shift(1) <= _df['BBL_5_2.0']).shift(1), 1, _df['pivot'])            
        elif tech_type == 'donchian':  # 
            plt.plot(_df.index, _df['DCU_20_20'], color='orange')  # Upper line
            plt.plot(_df.index, _df['DCL_20_20'], color='orange')  # Lower line
            # print(_df)
            # 분할매도는 상단기준선 넘으면서 바로 시작 (Not top) 해서 top 에서 매도 마무리
            _df['pivot'] = np.where((_df[tech_type] >= _df['DCU_20_20']) & (_df[tech_type].shift(1) <= _df['DCU_20_20']).shift(1), -2, 0)
            # 분할매도 마무리
            _df['pivot'] = np.where((_df[tech_type] >= _df['DCU_20_20']) & (_df[tech_type].shift(1) >= _df['DCU_20_20']).shift(1) & \
                                    (slope_before <= abs(permit_range)), -1, _df['pivot'])
            # 분할매수 시작은 Buttom 에서 시작해서 기준선 하단 통과시 매수 마무리
            _df['pivot'] = np.where((_df[tech_type] <= _df['DCL_20_20']) & (slope_before <= abs(permit_range)) & \
                                    (_df[tech_type].shift(1) <= _df['DCL_20_20']).shift(1), 2, _df['pivot'])
            # 분할매수 마무리
            _df['pivot'] = np.where((_df[tech_type] <= _df['DCL_20_20']) & (_df[tech_type].shift(1) <= _df['DCL_20_20']).shift(1), 1, _df['pivot'])  
        elif tech_type == 'kc':  # 
            plt.plot(_df.index, _df['KCUe_20_2'], color='orange')  # Upper line
            plt.plot(_df.index, _df['KCLe_20_2'], color='orange')  # Lower line
            # print(_df)
            # 분할매도는 상단기준선 넘으면서 바로 시작 (Not top) 해서 top 에서 매도 마무리
            _df['pivot'] = np.where((_df[tech_type] >= _df['KCUe_20_2']) & (_df[tech_type].shift(1) <= _df['KCUe_20_2']).shift(1), -2, 0)
            # 분할매도 마무리
            _df['pivot'] = np.where((_df[tech_type] >= _df['KCUe_20_2']) & (_df[tech_type].shift(1) >= _df['KCUe_20_2']).shift(1) & \
                                    (slope_before <= abs(permit_range)), -1, _df['pivot'])
            # 분할매수 시작은 Buttom 에서 시작해서 기준선 하단 통과시 매수 마무리
            _df['pivot'] = np.where((_df[tech_type] <= _df['KCLe_20_2']) & (slope_before <= abs(permit_range)) & \
                                    (_df[tech_type].shift(1) <= _df['KCLe_20_2']).shift(1), 2, _df['pivot'])
            # 분할매수 마무리
            _df['pivot'] = np.where((_df[tech_type] <= _df['KCLe_20_2']) & (_df[tech_type].shift(1) <= _df['KCLe_20_2']).shift(1), 1, _df['pivot'])  
        else:
            plt.plot(_df.index, _df[tech_type], color='orange')
            plt.axhline(y=h_limit, linestyle='--', lw=1.4, color='gray',)
            plt.axhline(y=l_limit, linestyle='--', lw=1.4, color='gray',)

            # Here is Buttom, 분할매수 시작, slope 이 마이너스(-)에서 0 로 전환된 날 리포트 : 2
            _df['pivot'] = np.where((slope_before <= abs(permit_range)) & (_df['SellAndBuy'].shift(1) < _df['SellAndBuy']) & \
                                    _df['SellAndBuy'] < 0, 2, 0)
            # 분할매수 끝, slope 이 플러스(+) 인 상태에서, 이동평균값이 기준선 0 위로 전환된 날 리포트 : 1 
            _df['pivot'] = np.where((slope_before > permit_range) & (_df['SellAndBuy'].shift(1) < _df['SellAndBuy']) & \
                                    _df['SellAndBuy'] < 0, 1, _df['pivot'])

            # Here is Top, 분할매도 시작, slope 이 플러스(+)에서 0 으로 전환된 날 리포트 : -2
            _df['pivot'] = np.where((slope_before <= abs(permit_range)) & (_df['SellAndBuy'].shift(1) > _df['SellAndBuy']) & \
                                    _df['SellAndBuy'] > 0, -2, _df['pivot'])
            # 분할매도 끝, slope 이 마이너스(-)인 상태에서 , 이동평균값이 기준선 0 아래로 전환된 날 리포트: -1
            _df['pivot'] = np.where((slope_before < permit_range*(-1)) & (_df['SellAndBuy'].shift(1) > _df['SellAndBuy']) & \
                                    _df['SellAndBuy'] > 0, -1, _df['pivot'])


            _df['pivot'] = np.where((_df[tech_type] >= h_limit) & (_df[tech_type].shift(1) <= h_limit), 1, 0)  # Sell signal
            _df['pivot'] = np.where((_df[tech_type] <= l_limit) & (_df[tech_type].shift(1) >= l_limit), -1, _df['pivot']) # Buy signal

        _1 = _df[_df['pivot'] == 2] # 분할매수 시작
        _2 = _df[_df['pivot'] == 1] # 분할매수 마무리
        _3 = _df[_df['pivot'] == -2] # 분할매도 시작
        _4 = _df[_df['pivot'] == -1] # 분할매도 마무리

        if not _1.empty:
            print(_1.iloc[-1]['pivot'])
        if not _2.empty:            
            print(_2.iloc[-1]['pivot'])
        if not _3.empty:            
            print(_3.iloc[-1]['pivot'])
        if not _4.empty:            
            print(_4.iloc[-1]['pivot'])

        for date in _1.index:
            plt.axvline(date, color='g', linestyle='--', lw=1,  alpha=1)
        # for date in _2.index:
        #     plt.axvline(date, color='g', linestyle='--', lw=1,  alpha=1)

        for date in _3.index:
            plt.axvline(date, color='r', linestyle='--', lw=1,  alpha=1) 
        # for date in _4.index:
        #     plt.axvline(date, color='r', linestyle='--', lw=1,  alpha=1)             
 
    signal = get_tech_types(tech_type)
    # print(signal)

    if _1.empty & _2.empty & _3.empty & _4.empty:
        latest_date = '2050-12-31' # 임시
        result = 0
        decision = '대기'
    else:
        if _1.empty & _2.empty & _3.empty:
            latest_date = _4.index[-1]
        elif _1.empty & _2.empty & _4.empty:
            latest_date = _3.index[-1]
        elif _1.empty & _3.empty & _4.empty:
            latest_date = _2.index[-1]
        elif _2.empty & _3.empty & _4.empty:
            latest_date = _1.index[-1]              
        elif _1.empty & _2.empty:
            latest_date = max(_3.index[-1], _4.index[-1])
        elif _1.empty & _3.empty:
            latest_date = max(_2.index[-1], _4.index[-1])
        elif _1.empty & _4.empty:
            latest_date = max(_2.index[-1], _3.index[-1]) 
        elif _2.empty & _3.empty:
            latest_date = max(_1.index[-1], _4.index[-1])
        elif _2.empty & _4.empty:
            latest_date = max(_1.index[-1], _3.index[-1])
        elif _1.empty:
            latest_date = max(_2.index[-1], _3.index[-1], _4.index[-1])
        elif _2.empty:
            latest_date = max(_1.index[-1], _3.index[-1], _4.index[-1])
        elif _3.empty:
            latest_date = max(_1.index[-1], _2.index[-1], _4.index[-1])
        elif _4.empty:
            latest_date = max(_1.index[-1], _2.index[-1], _3.index[-1])                
        else:
            latest_date = max(_1.index[-1], _2.index[-1], _3.index[-1], _4.index[-1])

        try:
            if latest_date == _1.index[-1]:  # 분할매수 시작
                result = 2
                if signal == 'trend':
                    decision = '추세상승, 분할매수 시작'
                elif signal == 'overBoughtSold':
                    decision = '과매도, 분할매수 시작'
                elif signal == 'trendReversal':
                    decision = '상승반전, 분할매수 시작'
                elif signal == 'kReversal':
                    decision = 'k 상승반전 분할매수 시작'
                else:
                    decision = '분할매수 시작'
        except IndexError as e:
            pass

        try:    
            if latest_date == _2.index[-1]:  # 분할매수 마무리
                result = 1
                if signal == 'trend':
                    decision = '상승추세 유지하지만 분할매수 마무리 '
                elif signal == 'overBoughtSold':
                    decision = '적정매도매수 분할매수 마무리'
                elif signal == 'trendReversal':
                    decision = '반전없음, 분할매수 마무리'
                elif signal == 'kReversal':
                    decision = 'k 반전없음, 분할매수 마무리'
                else:
                    decision = '분할매수 마무리'
        except IndexError as e:
            pass

        try:
            if latest_date == _3.index[-1]:  # 분할매도 시작
                result = -2
                if signal == 'trend':
                    decision = '추세하락, 분할매도 시작'
                elif signal == 'overBoughtSold':
                    decision = '과매수, 분할매도 시작'
                elif signal == 'trendReversal':
                    decision = '하락반정, 분할매도 시작'
                elif signal == 'kReversal':
                    decision = 'k 하락반전 분할매도 시작'
                else:
                    decision = '분할매도 시작'
        except IndexError as e:
            pass

        try:
            if latest_date == _4.index[-1]:  # 분할매도 마무리
                result = -1
                if signal == 'trend':
                    decision = '하락추세 유지하지만 분할매도 마무리'
                elif signal == 'overBoughtSold':
                    decision = '적정매수매도 분할매도 마무리'
                elif signal == 'trendReversal':
                    decision = '반전없음, 분할매도 마무리'
                elif signal == 'kReversal':
                    decision = 'k 반전없음, 분할매도 마무리'            
                else:
                    decision = '분할매도 마무리'
        except IndexError as e:
            pass

    plt.title(f"{ticker}: {tech_type} / {latest_date}부터 {decision} ", fontdict={'fontsize':20, 'color':'g'})    
    plt.grid(lw=0.7, color='lightgray')
    # plt.xlabel(f'Last Date: {latest_date}', loc='right', fontsize=18)
    
    return latest_date, result


'''
I. Technical Analysis

- 1. Trend
    Simple Moving Average (SMA)*
    Exponential Moving Average (EMA)*
    Weighted Moving Average (WMA)
    Moving Average Convergence Divergence (MACD)
    Average Directional Movement Index (ADX)*
    Vortex Indicator (VI)
    Trix (TRIX)
    Mass Index (MI)
    Detrended Price Oscillator (DPO)
    KST Oscillator (KST)
    Ichimoku Kinkō Hyō (Ichimoku)*
    Parabolic Stop And Reverse (Parabolic SAR)*
    Schaff Trend Cycle (STC)
    ZigZag Indicator.... (Not yet)

    
- 2. Momentum: https://medium.com/@crisvelasquez/top-6-momentum-indicators-in-python-bea0875e60a5
    Relative Strength Index (RSI)*
    Stochastic RSI (SRSI)
    True strength index (TSI)
    Ultimate Oscillator (UO)
    Stochastic Oscillator (SR)*
    Williams %R (WR)*
    Awesome Oscillator (AO)*
    Kaufman’s Adaptive Moving Average (KAMA)
    Rate of Change (ROC)*
    Percentage Price Oscillator (PPO)
    Percentage Volume Oscillator (PVO)
    Commodity Channel Index (CCI)*

    
- 3. Volume
    Money Flow Index (MFI)*
    Accumulation/Distribution Index (ADI)*
    On-Balance Volume (OBV)*
    Volume-price Trend (VPT)*    
    Chaikin Money Flow (CMF)*
    Force Index (FI)
    Ease of Movement (EoM, EMV)
    Negative Volume Index (NVI)*
    Volume Weighted Average Price (VWAP)*
    Ulcer Index (UI)
    Positive Volume Index (PVI)*
    Klinger Volume Oscillator(KVO)*

    
- 4. Volatility
    Average True Range (ATR)*
    Bollinger Bands (BB)*
    Donchian Channel (DC)*
    Keltner Channel (KC)*
    Relative Volatility Index (RVI).... (Not yet)
    Volatility Chaikin.... (Not yet)

'''

'''
1. Trend Indicators: https://medium.com/@crisvelasquez/top-6-trend-indicators-in-python-c922ac0674f9
'''
# Trend.Simple Moving Average (SMA)
# calculates the average price over a specific period, providing a smoothed representation of price movement.
# 특정 기간 동안의 평균 가격을 계산하여 가격 변동을 부드럽게 표현합니다.
def sma(ticker, df, tech_type, TECH_LIST, idx):
    try:
        df[tech_type] = df.ta.sma(20) - df.ta.sma(200)
    except ValueError as e:  # 530107.KS: history 건수가 매우 적어 20일 또는 200일 이동평균을 구하지 못하는 오류
        base.logger2.error(f'{ticker} Technical.sma Value error 1: {e}')
        return df, '2050-12-31', 0, 0
    
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)

    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Trend.Exponential Moving Average (EMA)
# gives more weight to recent prices, making it more responsive to new information.
# 최근 가격에 더 많은 가중치를 부여하여 새로운 정보에 더 잘 반응합니다.
def ema(ticker, df, tech_type, TECH_LIST, idx):
    try:
        df[tech_type] = df.ta.ema(20) - df.ta.ema(200)
    except ValueError as e:  # 530107.KS: history 건수가 매우 적어 20일 또는 200일 이동평균을 구하지 못하는 오류
        base.logger2.error(f'{ticker} Technical.ema Value error 1: {e}')
        return df, '2050-12-31', 0, 0

    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Trend.Exponential Moving Average (EMA)
# shows the relationship between two moving averages of a security’s price.
# 증권 가격의 두 이동 평균 사이의 관계를 보여줍니다.
def macd(ticker, df, tech_type, TECH_LIST, idx):
    # Get the 26-day EMA of the closing price
    k = df['Close'].ewm(span=12, adjust=False, min_periods=12).mean()
    # Get the 12-day EMA of the closing price
    d = df['Close'].ewm(span=26, adjust=False, min_periods=26).mean()
    # Subtract the 26-day EMA from the 12-Day EMA to get the MACD
    macd = k - d
    # Get the 9-Day EMA of the MACD for the Trigger line
    macd_s = macd.ewm(span=9, adjust=False, min_periods=9).mean()
    # Calculate the difference between the MACD - Trigger for the Convergence/Divergence value
    df[tech_type] = (macd - macd_s)*5  # 폭이 너무 작아 크게 보이기 위해 *5
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Average Directional Movement Index (ADX):
# quantify the strength of a trend. 추세의 강도를 정량화합니다.
# The +DI and -DI lines can help identify the direction of the trend. 양의 방향 표시기(+DI) 및 음의 방향 표시기(-DI).
def adx(ticker, df, tech_type, TECH_LIST, idx):  # DMP_14(posituve) 와 DMN_14(negative) 두 개 칼럼 생성
    df.ta.adx(length=14, append=True)
    df[tech_type] = df['ADX_14']
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Parabolic Stop And Reverse (Parabolic SAR)
def psar(ticker, df, tech_type, TECH_LIST, idx):  # DMP_14(posituve) 와 DMN_14(negative) 두 개 칼럼 생성
    df.ta.psar(append=True)
    df['PSARl_0.02_0.2'] = df['PSARl_0.02_0.2'].fillna(0)
    df['PSARs_0.02_0.2'] = df['PSARs_0.02_0.2'].fillna(0)
    df[tech_type] = df['Close'] - (df['PSARl_0.02_0.2'] + df['PSARs_0.02_0.2'])  # PSARl_0.02_0.2 는 long 값, PSARl_0.02_0.2 는 short 값
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy


# Ichimoku Kinkō Hyō (Ichimoku)
# a trend-following indicator that provides entry and exit points.
# 진입점과 이탈점을 제공하는 추세추종 지표입니다.
def ichmoku(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.ichimoku(append=True)
    # Tenkan Sen: ITS_9 > Kijun Sen: IKS_26 => Buy signal
    df[tech_type] = (df['ITS_9'] - df['IKS_26'])*5  # 폭이 너무 작아 크게 보이기 위해 *5
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy


'''
2. Momentum Indicators: https://medium.com/@crisvelasquez/top-6-momentum-indicators-in-python-bea0875e60a5
'''
# Momentum.Relative Strength Index (RSI) 
# measures the speed and change of price movements, oscillating between 0 and 100.
# 0과 100 사이에서 진동하는 가격 움직임의 속도와 변화를 측정합니다.
def rsi(ticker, df, tech_type, TECH_LIST, idx):    
    df[tech_type] = df.ta.rsi(14)
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Stochastic Oscillator (STOCH): 
# comparing a particular closing price of a security to a range of its prices over a certain period of time.
# 특정 기간 동안의 특정 유가 증권 종가를 해당 가격 범위와 비교하는 것입니다.
def stoch(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.stoch(append=True)
    df[tech_type] = df['STOCHk_14_3_3'] - df['STOCHd_14_3_3']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Rate of Change (ROC) Momentum Oscillator
# measures the percentage change in price between the current price and the price a certain number of periods ago.
# 현재 가격과 특정 기간 전 가격 사이의 가격 변화율을 측정합니다.
def roc(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.roc(append=True)
    df[tech_type] = df['ROC_10']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Commodity Channel Index (CCI): identifies cyclical trends in securities 
# by comparing their current typical price (TP) to the average TP over a specific period, usually 20 days.
# 현재 일반 가격(TP)을 특정 기간(보통 20일) 동안의 평균 TP와 비교하여 증권의 순환 추세를 식별합니다.
def cci(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.cci(append=True)
    df[tech_type] = df['CCI_14_0.015']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Williams %R (WR): reflects the level of the close relative to the highest high for a set period.
# 일정 기간 동안 최고가 대비 종가 수준을 반영합니다.
def willr(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.willr(append=True)
    df[tech_type] = df['WILLR_14']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Awesome Oscillator (AO): designed to capture momentum in the market by comparing the recent market momentum with the general momentum over a wider frame of time.
# 최근 시장 모멘텀과 더 넓은 기간 동안의 일반적인 모멘텀을 비교하여 시장의 모멘텀을 포착합니다.
def ao(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.ao(append=True)
    df[tech_type] = df['AO_5_34']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10

    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Stochastic RSI (STOCHRSI)
def stochrsi(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.stochrsi(append=True)    
    df[tech_type] = df['STOCHk_14_3_3'] - df['STOCHd_14_3_3']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Momentum.Percentage Price Oscillator (PPO):
# "Stochastic RSI and Dynamic Momentum Index" was created by Tushar Chande and Stanley Kroll and published in Stock & Commodities V.11:5 (189-199)
# It is a range-bound oscillator with two lines moving between 0 and 100.
def ppo(ticker, df, tech_type, TECH_LIST, idx):    
    buf= df.ta.ppo(close=df['Close'], append=True)
    df[tech_type] = buf['PPO_12_26_9']
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
                    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

'''
3. Volume Indicators: https://medium.com/@crisvelasquez/top-9-volume-indicators-in-python-e398791b98f9
'''
# Volume.On-Balance Volume (OBV)
# a cumulative indicator that relates volume to price change. 
# OBV increases or decreases during each trading day in line with whether the price closes higher or lower from the previous close.
# 거래량과 가격 변화를 연결하는 누적 지표입니다.
# OBV는 가격이 이전 종가보다 높게 또는 낮게 마감되는지 여부에 따라 각 거래일 동안 증가하거나 감소합니다.
def obv(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.obv(append=True)
    df['OBV_EMA'] = df['OBV'].ewm(span=30).mean()  # 20-day EMA of OBV
    df[tech_type] = df['OBV'] - df['OBV_EMA']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Volume Price Trend (VPT)
# integrates volume with the percentage change in price, cumulatively indicating the trend’s strength and direction.
# 거래량과 가격 변동률을 통합하여 추세의 강도와 방향을 누적적으로 나타냅니다.
def pvt(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.pvt(append=True)
    df['PVT_MA'] = df['PVT'].rolling(window=30).mean()  # 20-day moving average
    df[tech_type] = df['PVT'] - df['PVT_MA']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Positive Volume Index (PVI)
# @@@quantifies the percentage rate at which volume changes over a specific period, comparing current volume to volume from n days ago.
# @@@n일 전의 거래량과 현재 거래량을 비교하여 특정 기간 동안 거래량이 변경되는 비율을 정량화합니다.
def pvi(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.pvi(append=True)
    df['PVI_1_MA'] = df['PVI_1'].rolling(window=30).mean()  # 20-day moving average
    df[tech_type] = df['PVI_1'] - df['PVI_1_MA']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Chaikin Money Flow (CMF)
# integrates price and volume to measure the buying and selling pressure over a specified period, typically 20 days.
# 특정 기간(일반적으로 20일) 동안의 매수 및 매도 압력을 측정하기 위해 가격과 거래량을 통합합니다.
def cmf(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.cmf(append=True)
    df[tech_type] = df['CMF_20']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Volume-Weighted Average Price (VWAP)
# gives the average price a stock has traded at during the day, weighted by volume
# 거래량에 따라 가중치를 적용하여 하루 동안 주식이 거래된 평균 가격을 제공합니다.
def vwap(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.vwap(append=True)
    df[tech_type] = df['Close'] - df['VWAP_D']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Accumulation/Distribution Oscillator (ADOSC)
# designed to reflect the cumulative flow of money into or out of a security, factoring in both the volume and the price movement.
# 거래량과 가격 변동을 모두 고려하여 증권 안팎으로의 누적 자금 흐름을 반영하도록 설계되었습니다.
def adosc(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.adosc(append=True)
    df[tech_type] = df['ADOSC_3_10']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Money Flow Index (MFI)
# a technical momentum indicator that combines price and volume to assess the buying or selling pressure on an asset. 
# 자산에 대한 매수 또는 매도 압력을 평가하기 위해 가격과 거래량을 결합하는 기술적 모멘텀 지표입니다.
def mfi(ticker, df, tech_type, TECH_LIST, idx):
    df[tech_type] = df.ta.mfi(close=df['Close'],length=14, append=True)
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)

    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Klinger Volume Oscillator (KVO)
# It is designed to predict price reversals in a market by comparing volume to price
# 거래량과 가격을 비교하여 시장의 가격 반전을 예측하도록 설계되었습니다.
def kvo(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.kvo(append=True)
    df[tech_type] = df['KVO_34_55_13']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volume.Negative Volume Index (NVI)
# cumulative indicator that increases when the volume decreases compared to the previous session, suggesting that the “smart money” is active.
# 이전 세션에 비해 거래량이 감소할 때 증가하는 누적 지표로 '스마트머니'가 활성화되었음을 시사한다.
def nvi(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.nvi(append=True)
    df['NVI_1_SMA'] = df['NVI_1'].rolling(window=20).mean()    
    df[tech_type] = df['NVI_1'] - df['NVI_1_SMA']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy


'''
4. Volatility Indicators: https://medium.com/@crisvelasquez/top-6-volatility-indicators-in-python-14fbd7bf92d8
'''
# Volatility.Average True Range (ATR)
# quantifies market volatility by averaging the range of price movements.
# 가격 변동 범위를 평균하여 시장 변동성을 정량화합니다.
def atr(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.atr(append=True)
    df['ATRr_14_MA'] = df['ATRr_14'].rolling(window=20).mean()    
    df[tech_type] = df['ATRr_14'] - df['ATRr_14_MA']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volatility.Bollinger Bands (BB)
# provide insights into market volatility and overbought/oversold conditions.
# 시장 변동성과 과매수/과매도 상태에 대한 통찰력을 제공합니다.
def bbands(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.bbands(append=True)
    df[tech_type] = df['Close']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)    
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volatility.Donchian Channels (DONCHAIN)
# are based on the highest high and lowest low, offering a view of market range and volatility.
# 최고가와 최저가를 기반으로 하며 시장 범위와 변동성에 대한 시각을 제공합니다.
def donchian(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.donchian(append=True)
    df[tech_type] = df['Close']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)    
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volatility.Keltner Channels (KC)
# use ATR for band setting, making them sensitive to volatility spikes.
# 밴드 설정에 ATR을 사용하여 변동성 급증에 민감하게 만듭니다.
def kc(ticker, df, tech_type, TECH_LIST, idx):    
    df.ta.kc(append=True)
    df[tech_type] = df['Close']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)    
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy

# Volatility.Relative Volatility Index (RVI)
# volatility measurement that mirrors the RSI but uses standard deviation.
# RSI를 반영하지만 표준편차를 사용하는 변동성 측정입니다.
def rvi(ticker, df, tech_type, TECH_LIST, idx):
    df.ta.rvi(append=True)
    df['RVI_14_SMA'] = df['RVI_14'].rolling(window=20).mean()    
    df[tech_type] = df['RVI_14'] - df['RVI_14_SMA']
    # print(df.tail())
    latest_date, sellandbuy = make_tech_plot(ticker, df, tech_type, TECH_LIST, idx)
    
    try:
        from_latest_date_slope = base.trend_detector(df, tech_type, latest_date)
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_latest_date_slope = (df[tech_type] - df[tech_type].shift(10)) / 10
    try:
        from_5day_ago_slope = base.trend_detector(df, tech_type, base.get_working_day_before(5))
    except Exception as e:  # lastest_date 가 짧아 np.polyfit 함수 오류로 인한 수작업 기울기 구하기.
        from_5day_ago_slope = (df[tech_type] - df[tech_type].shift(3)) / 3 # 더 빠른 반응을 위해. 

    slope = (from_latest_date_slope / from_5day_ago_slope) * 5
    
    return df, latest_date, slope, sellandbuy